/* Copyright © 2013, Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#define _GNU_SOURCE
#include <dlfcn.h>

#include <stdio.h>
#include <stdlib.h>

void *libgl_handle;
void *wrapper_handle;

static void
open_libgl_handle (void)
{
	const char *path;

	if (libgl_handle)
		return;

	path = getenv ("GLAZE_LIBGL");
	if (path == NULL) {
		fprintf (stderr, "GLAZE_LIBGL unset. Please set to path of real libGL.so under glaze.\n");
		exit (1);
	}

	libgl_handle = dlopen (path, RTLD_LAZY | RTLD_GLOBAL);
	if (libgl_handle == NULL) {
		fprintf (stderr, "Error: Failed to dlopen %s\n", path);
		exit (1);
	}
}

static void
open_wrapper_handle (void)
{
	const char *path;

	if (wrapper_handle)
		return;

	path = getenv ("GLAZE_WRAPPER");
	if (path == NULL) {
		fprintf (stderr, "GLAZE_WRAPPER unset. Please set to path of real libGL.so under glaze.\n");
		exit (1);
	}

	wrapper_handle = dlopen (path, RTLD_LAZY);
	if (wrapper_handle == NULL) {
		fprintf (stderr, "Error: Failed to dlopen %s\n", path);
		exit (1);
	}
}

static void
glaze_init (void) __attribute__((constructor));

static void
glaze_init (void)
{
	open_libgl_handle ();
	open_wrapper_handle ();
}

static void *
resolve (const char *name)
{
	void *symbol;

	/* The wrapper gets first choice on all symbols. */
	symbol = dlsym (wrapper_handle, name);
	if (symbol)
		return symbol;

	return dlsym (libgl_handle, name);
}

#define GLAZE_API(name)							\
void * name() __attribute__((ifunc(#name "_resolver")));		\
static void *								\
name ## _resolver (void) { return resolve (#name); }

#include "specs/gl.def"
#include "specs/glx.def"
